# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

cimport numpy


cdef extern from "distributions/vector.hpp" namespace "distributions":
    cdef cppclass VectorFloat:
        cppclass iterator:
            float& operator*() nogil
            iterator operator++() nogil
            iterator operator--() nogil
            iterator operator+(size_t) nogil
            iterator operator-(size_t) nogil
            bint operator==(iterator) nogil
            bint operator!=(iterator) nogil
            bint operator<(iterator) nogil
            bint operator>(iterator) nogil
            bint operator<=(iterator) nogil
            bint operator>=(iterator) nogil
        cppclass reverse_iterator:
            float& operator*() nogil
            iterator operator++() nogil
            iterator operator--() nogil
            iterator operator+(size_t) nogil
            iterator operator-(size_t) nogil
            bint operator==(reverse_iterator) nogil
            bint operator!=(reverse_iterator) nogil
            bint operator<(reverse_iterator) nogil
            bint operator>(reverse_iterator) nogil
            bint operator<=(reverse_iterator) nogil
            bint operator>=(reverse_iterator) nogil
        VectorFloat() nogil except +
        VectorFloat(VectorFloat&) nogil except +
        VectorFloat(size_t) nogil except +
        VectorFloat(size_t, float&) nogil except +
        float& operator[](size_t) nogil
        bint operator==(VectorFloat&, VectorFloat&) nogil
        bint operator!=(VectorFloat&, VectorFloat&) nogil
        bint operator<(VectorFloat&, VectorFloat&) nogil
        bint operator>(VectorFloat&, VectorFloat&) nogil
        bint operator<=(VectorFloat&, VectorFloat&) nogil
        bint operator>=(VectorFloat&, VectorFloat&) nogil
        void assign(size_t, float&) nogil
        float& at(size_t) nogil
        float& back() nogil
        iterator begin() nogil
        size_t capacity() nogil
        void clear() nogil
        bint empty() nogil
        iterator end() nogil
        iterator erase(iterator) nogil
        iterator erase(iterator, iterator) nogil
        float& front() nogil
        iterator insert(iterator, float&) nogil
        void insert(iterator, size_t, float&) nogil
        void insert(iterator, iterator, iterator) nogil
        size_t max_size() nogil
        void pop_back() nogil
        void push_back(float&) nogil
        reverse_iterator rbegin() nogil
        reverse_iterator rend() nogil
        void reserve(size_t) nogil
        void resize(size_t) nogil
        void resize(size_t, float&) nogil
        size_t size() nogil
        void swap(VectorFloat&) nogil
        float * data() nogil

    cdef cppclass AlignedFloats:
        AlignedFloats (float *, size_t) nogil
        float * data () nogil
        size_t size () nogil


cdef void vector_float_from_ndarray(
        VectorFloat & vector_float,
        numpy.ndarray[numpy.float32_t, ndim=1] ndarray)


cdef void vector_float_to_ndarray(
        VectorFloat & vector_float,
        numpy.ndarray[numpy.float32_t, ndim=1] ndarray)



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import glob
from collections import defaultdict
from itertools import izip
import math
import numpy
from numpy.testing import assert_array_almost_equal
from nose import SkipTest
from nose.tools import assert_true, assert_less, assert_equal
import importlib
import distributions
from distributions.util import multinomial_goodness_of_fit

ROOT = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
TOL = 1e-3


def require_cython():
    if not distributions.has_cython:
        raise SkipTest('no cython support')


def seed_all(s):
    import distributions.dbg.random
    distributions.dbg.random.seed(s)
    try:
        import distributions.hp.random
        distributions.hp.random.seed(s)
    except ImportError:
        pass


def list_models():
    result = set()
    for path in glob.glob(os.path.join(ROOT, '*', 'models', '*.p*')):
        dirname, filename = os.path.split(path)
        flavor = os.path.split(os.path.dirname(dirname))[-1]
        name = os.path.splitext(filename)[0]
        if not name.startswith('__'):
            result.add((name, flavor))
    for name, flavor in sorted(result):
        spec = {'flavor': flavor, 'name': name}
        if name.startswith('_'):
            continue
        try:
            import_model(spec)
            yield spec
        except ImportError:
            module_name = 'distributions.{flavor}.models.{name}'.format(**spec)
            print 'failed to import {}'.format(module_name)


def import_model(spec):
    module_name = 'distributions.{flavor}.models.{name}'.format(**spec)
    return importlib.import_module(module_name)


def assert_hasattr(thing, attr):
    assert_true(
        hasattr(thing, attr),
        "{} is missing attribute '{}'".format(thing.__name__, attr))


def print_short(x, size=64):
    string = str(x)
    if len(string) > size:
        string = string[:size - 3] + '...'
    return string


def assert_close(lhs, rhs, tol=TOL, err_msg=None):
    try:
        if isinstance(lhs, dict):
            assert_true(
                isinstance(rhs, dict),
                'type mismatch: {} vs {}'.format(type(lhs), type(rhs)))
            assert_equal(set(lhs.keys()), set(rhs.keys()))
            for key, val in lhs.iteritems():
                msg = '{}[{}]'.format(err_msg or '', key)
                assert_close(val, rhs[key], tol, msg)
        elif isinstance(lhs, float) or isinstance(lhs, numpy.float64):
            assert_true(
                isinstance(rhs, float) or isinstance(rhs, numpy.float64),
                'type mismatch: {} vs {}'.format(type(lhs), type(rhs)))
            diff = abs(lhs - rhs)
            norm = 1 + abs(lhs) + abs(rhs)
            msg = '{} off by {}% = {}'.format(
                err_msg or '',
                100 * diff / norm,
                diff)
            assert_less(diff, tol * norm, msg)
        elif isinstance(lhs, numpy.ndarray) or isinstance(rhs, numpy.ndarray):
            assert_true(
                (isinstance(lhs, numpy.ndarray) or isinstance(lhs, list)) and
                (isinstance(rhs, numpy.ndarray) or isinstance(rhs, list)),
                'type mismatch: {} vs {}'.format(type(lhs), type(rhs)))
            decimal = int(round(-math.log10(tol)))
            assert_array_almost_equal(
                lhs,
                rhs,
                decimal=decimal,
                err_msg=(err_msg or ''))
        elif isinstance(lhs, list) or isinstance(lhs, tuple):
            assert_true(
                isinstance(rhs, list) or isinstance(rhs, tuple),
                'type mismatch: {} vs {}'.format(type(lhs), type(rhs)))
            for pos, (x, y) in enumerate(izip(lhs, rhs)):
                msg = '{}[{}]'.format(err_msg or '', pos)
                assert_close(x, y, tol, msg)
        else:
            assert_equal(lhs, rhs, err_msg)
    except Exception:
        print err_msg or ''
        print 'actual = {}'.format(print_short(lhs))
        print 'expected = {}'.format(print_short(rhs))
        raise


def assert_all_close(collection, **kwargs):
    for i1, item1 in enumerate(collection[:-1]):
        for item2 in collection[i1 + 1:]:
            assert_close(item1, item2, **kwargs)


def collect_samples_and_scores(sampler, total_count=10000):
    '''
    Collect samples and MC estimates of sample probabilities.

    Inputs:
        - sampler generates (sample, prob) pairs.  samples must be hashable.
          probs may be randomized, but must be unbiased and low-variance.
        - total_count samples are drawn in total.
        - tol is the minimum goodness of fit allowed to pass the test.
    Returns:
        - counts : key -> int
        - probs : key -> float
    '''
    counts = defaultdict(lambda: 0)
    probs = defaultdict(lambda: 0.0)
    for _ in xrange(total_count):
        sample, prob = sampler()
        counts[sample] += 1
        probs[sample] += prob

    for key, count in counts.iteritems():
        probs[key] /= count
    total_prob = sum(probs.itervalues())
    assert_close(total_prob, 1.0, tol=1e-2, err_msg='total_prob is biased')

    return counts, probs


def assert_counts_match_probs(counts, probs, tol=1e-3):
    '''
    Check goodness of fit of observed counts to predicted probabilities
    using Pearson's chi-squared test.

    Inputs:
        - counts : key -> int
        - probs : key -> float
    '''
    keys = counts.keys()
    probs = [probs[key] for key in keys]
    counts = [counts[key] for key in keys]
    total_count = sum(counts)

    print 'EXPECT\tACTUAL\tVALUE'
    for prob, count, key in sorted(izip(probs, counts, keys), reverse=True):
        expect = prob * total_count
        print '{:0.1f}\t{}\t{}'.format(expect, count, key)

    gof = multinomial_goodness_of_fit(probs, counts, total_count)
    print 'goodness of fit = {}'.format(gof)
    assert gof > tol, 'failed with goodness of fit {}'.format(gof)


def assert_samples_match_scores(sampler, total_count=10000, tol=1e-3):
    '''
    Test that a discrete sampler is distributed according to its scores.

    Inputs:
        - sampler generates (sample, prob) pairs.  samples must be hashable.
          probs may be randomized, but must be unbiased and low-variance.
        - total_count samples are drawn in total.
        - tol is the minimum goodness of fit allowed to pass the test.
    '''
    counts, probs = collect_samples_and_scores(sampler, total_count)
    assert_counts_match_probs(counts, probs)



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

cimport _nich
import _nich

from distributions.mixins import SharedMixin, GroupIoMixin, SharedIoMixin


NAME = 'NormalInverseChiSq'
EXAMPLES = [
    {
        'shared': {'mu': 0., 'kappa': 1., 'sigmasq': 1., 'nu': 1.},
        'values': [-4.0, -2.0, -1.0, -0.5, 0.0, 0.5, 1.0, 2.0, 4.0],
    },
]
Value = float


cdef class _Shared(_nich.Shared):
    def load(self, dict raw):
        self.ptr.mu = raw['mu']
        self.ptr.kappa = raw['kappa']
        self.ptr.sigmasq = raw['sigmasq']
        self.ptr.nu = raw['nu']

    def dump(self):
        return {
            'mu': self.ptr.mu,
            'kappa': self.ptr.kappa,
            'sigmasq': self.ptr.sigmasq,
            'nu': self.ptr.nu,
        }

    def load_protobuf(self, message):
        self.ptr.mu = message.mu
        self.ptr.kappa = message.kappa
        self.ptr.sigmasq = message.sigmasq
        self.ptr.nu = message.nu

    def dump_protobuf(self, message):
        message.Clear()
        message.mu = self.ptr.mu
        message.kappa = self.ptr.kappa
        message.sigmasq = self.ptr.sigmasq
        message.nu = self.ptr.nu


class Shared(_Shared, SharedMixin, SharedIoMixin):
    pass


cdef class _Group(_nich.Group):
    def load(self, dict raw):
        self.ptr.count = raw['count']
        self.ptr.mean = raw['mean']
        self.ptr.count_times_variance = raw['count_times_variance']

    def dump(self):
        return {
            'count': self.ptr.count,
            'mean': self.ptr.mean,
            'count_times_variance': self.ptr.count_times_variance,
        }


class Group(_Group, GroupIoMixin):
    pass


class Sampler(_nich.Sampler):
    pass


Mixture = _nich.Mixture
sample_group = _nich.sample_group



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from distributions.dbg.special import log, gammaln
from distributions.dbg.random import sample_bernoulli, sample_beta
from distributions.mixins import SharedMixin, GroupIoMixin, SharedIoMixin


NAME = 'BetaBernoulli'
EXAMPLES = [
    {
        'shared': {'alpha': 0.5, 'beta': 2.0},
        'values': [False, False, True, False, True, True, False, False],
    },
    {
        'shared': {'alpha': 10.5, 'beta': 0.5},
        'values': [False, False, False, False, False, False, False, True],
    },
]
Value = bool


class Shared(SharedMixin, SharedIoMixin):
    def __init__(self):
        self.alpha = None
        self.beta = None

    def load(self, raw):
        self.alpha = float(raw['alpha'])
        self.beta = float(raw['beta'])

    def dump(self):
        return {
            'alpha': self.alpha,
            'beta': self.beta,
        }

    def load_protobuf(self, message):
        self.alpha = float(message.alpha)
        self.beta = float(message.beta)

    def dump_protobuf(self, message):
        message.alpha = self.alpha
        message.beta = self.beta


class Group(GroupIoMixin):
    def __init__(self):
        self.heads = None
        self.tails = None

    def init(self, shared):
        self.heads = 0
        self.tails = 0

    def add_value(self, shared, value):
        if value:
            self.heads += 1
        else:
            self.tails += 1

    def add_repeated_value(self, shared, value, count):
        if value:
            self.heads += count
        else:
            self.tails += count

    def remove_value(self, shared, value):
        if value:
            self.heads -= 1
        else:
            self.tails -= 1

    def merge(self, shared, source):
        self.heads += source.heads
        self.tails += source.tails

    def score_value(self, shared, value):
        """
        \cite{wallach2009rethinking} Eqn 4.
        McCallum, et. al, 'Rething LDA: Why Priors Matter'
        """
        heads = shared.alpha + self.heads
        tails = shared.beta + self.tails
        numer = heads if value else tails
        denom = heads + tails
        return log(numer / denom)

    def score_data(self, shared):
        """
        \cite{jordan2001more} Eqn 22.
        Michael Jordan's CS281B/Stat241B
        Advanced Topics in Learning and Decision Making course,
        'More on Marginal Likelihood'
        """
        alpha = shared.alpha + self.heads
        beta = shared.beta + self.tails
        score = gammaln(shared.alpha + shared.beta) - gammaln(alpha + beta)
        score += gammaln(alpha) - gammaln(shared.alpha)
        score += gammaln(beta) - gammaln(shared.beta)
        return score

    def sample_value(self, shared):
        sampler = Sampler()
        sampler.init(shared, self)
        return sampler.eval(shared)

    def load(self, raw):
        self.heads = raw['heads']
        self.tails = raw['tails']

    def dump(self):
        return {
            'heads': self.heads,
            'tails': self.tails,
        }

    def load_protobuf(self, message):
        self.heads = message.heads
        self.tails = message.tails

    def dump_protobuf(self, message):
        message.heads = self.heads
        message.tails = self.tails


class Sampler(object):
    def init(self, shared, group=None):
        if group is None:
            self.p = sample_beta(shared.alpha, shared.beta)
        else:
            alpha = shared.alpha + group.heads
            beta = shared.beta + group.tails
            self.p = sample_beta(alpha, beta)

    def eval(self, shared):
        return sample_bernoulli(self.p)


def sample_group(shared, size):
    group = Group()
    group.init(shared)
    sampler = Sampler()
    sampler.init(shared, group)
    return [sampler.eval(shared) for _ in xrange(size)]



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import bz2
import gzip
import simplejson
import struct


def open_compressed(filename, mode='r'):
    if filename.endswith('.bz2'):
        return bz2.BZ2File(filename, mode.replace('b', ''))
    elif filename.endswith('.gz'):
        return gzip.GzipFile(filename, mode)
    else:
        return file(filename, mode)


def json_dump(data, filename, **kwargs):
    dirname = os.path.dirname(filename)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)
    with open_compressed(filename, 'w') as f:
        simplejson.dump(data, f, **kwargs)


def json_load(filename):
    with open_compressed(filename, 'rb') as f:
        return simplejson.load(f)


def json_stream_dump(stream, filename, **kwargs):
    kwargs['separators'] = (',', ':')
    stream = iter(stream)
    dirname = os.path.dirname(filename)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)
    with open_compressed(filename, 'w') as f:
        f.write('[')
        try:
            item = next(stream)
            f.write('\n')
            simplejson.dump(item, f, **kwargs)
            for item in stream:
                f.write(',\n')
                simplejson.dump(item, f, **kwargs)
        except StopIteration:
            pass
        f.write('\n]')


def json_costream_dump(filename, **kwargs):
    kwargs['separators'] = (',', ':')
    dirname = os.path.dirname(filename)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)
    with open_compressed(filename, 'w') as f:
        f.write('[')
        try:
            item = (yield)
            f.write('\n')
            simplejson.dump(item, f, **kwargs)
            while True:
                item = (yield)
                f.write(',\n')
                simplejson.dump(item, f, **kwargs)
        except GeneratorExit:
            pass
        f.write('\n]')


class json_stream_load(object):
    '''
    Read json data that was created by json_stream_dump or json_costream_dump.

    Note that this exploits newline formatting in the above dumpers.
    In particular:
    - the first line is '['
    - intermediate lines are of the form '{},'.format(json_parsable_content)
    - the penultimate line is of the form '{}'.format(json_parsable_content)
    - the last line is ']'
    - there is no trailing whitespace

    An alternative would be to use ijson to streamingly load arbitrary json
    files, however in practice this is ~40x slower.
    '''
    def __init__(self, filename):
        self.fd = open_compressed(filename, 'rb')
        line = self.fd.readline(2)
        if line != '[\n':
            raise IOError(
                'Unhandled format for json_stream_load. '
                'Try recreating json file with the compatible '
                'json_stream_dump or json_costream_dump.')

    def __iter__(self):
        return self

    def next(self):
        line = self.fd.readline().rstrip(',\n')
        if line == ']':
            self.close()
            raise StopIteration
        else:
            return simplejson.loads(line)

    def close(self):
        self.fd.close()


def protobuf_stream_write(item, fd):
    assert isinstance(item, str), item
    fd.write(struct.pack('<I', len(item)))
    fd.write(item)


def protobuf_stream_read(fd):
    size_str = fd.read(4)
    if len(size_str) < 4:
        raise StopIteration
    size = struct.unpack('<I', size_str)[0]
    return fd.read(size)


def protobuf_stream_dump(stream, filename):
    dirname = os.path.dirname(filename)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)
    with open_compressed(filename, 'wb') as f:
        for item in stream:
            protobuf_stream_write(item, f)


class protobuf_stream_load(object):
    def __init__(self, filename):
        self.fd = open_compressed(filename, 'rb')

    def __iter__(self):
        return self

    def next(self):
        return protobuf_stream_read(self.fd)

    def close(self):
        self.fd.close()



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from distributions.dbg.special import log, factorial, gammaln
from distributions.dbg.random import sample_gamma, sample_poisson
from distributions.mixins import SharedMixin, GroupIoMixin, SharedIoMixin


NAME = 'GammaPoisson'
EXAMPLES = [
    {
        'shared': {'alpha': 1., 'inv_beta': 1.},
        'values': [0, 1, 2, 3, 4, 5, 6, 1, 2, 3, 4, 2, 3],
    }
]
Value = int


class Shared(SharedMixin, SharedIoMixin):
    def __init__(self):
        self.alpha = None
        self.inv_beta = None

    def plus_group(self, group):
        post = self.__class__()
        post.alpha = self.alpha + group.sum
        post.inv_beta = self.inv_beta + group.count
        return post

    def load(self, raw):
        self.alpha = float(raw['alpha'])
        self.inv_beta = float(raw['inv_beta'])

    def dump(self):
        return {
            'alpha': self.alpha,
            'inv_beta': self.inv_beta,
        }

    def load_protobuf(self, message):
        self.alpha = float(message.alpha)
        self.inv_beta = float(message.inv_beta)

    def dump_protobuf(self, message):
        message.Clear()
        message.alpha = self.alpha
        message.inv_beta = self.inv_beta


class Group(GroupIoMixin):
    def __init__(self):
        self.count = None
        self.sum = None
        self.log_prod = None

    def init(self, shared):
        self.count = 0
        self.sum = 0
        self.log_prod = 0.

    def add_value(self, shared, value):
        self.count += 1
        self.sum += int(value)
        self.log_prod += log(factorial(value))

    def add_repeated_value(self, shared, value, count):
        self.count += count
        self.sum += int(count * value)
        self.log_prod += count * log(factorial(value))

    def remove_value(self, shared, value):
        self.count -= 1
        self.sum -= int(value)
        self.log_prod -= log(factorial(value))

    def merge(self, shared, source):
        self.count += source.count
        self.sum += source.sum
        self.log_prod += source.log_prod

    def score_value(self, shared, value):
        post = shared.plus_group(self)
        return gammaln(post.alpha + value) - gammaln(post.alpha) \
            + post.alpha * log(post.inv_beta) \
            - (post.alpha + value) * log(1. + post.inv_beta) \
            - log(factorial(value))

    def score_data(self, shared):
        post = shared.plus_group(self)
        return gammaln(post.alpha) - gammaln(shared.alpha) \
            - post.alpha * log(post.inv_beta) \
            + shared.alpha * log(shared.inv_beta) \
            - self.log_prod

    def sample_value(self, shared):
        sampler = Sampler()
        sampler.init(shared, self)
        return sampler.eval(shared)

    def load(self, raw):
        self.count = int(raw['count'])
        self.sum = int(raw['sum'])
        self.log_prod = float(raw['log_prod'])

    def dump(self):
        return {
            'count': self.count,
            'sum': self.sum,
            'log_prod': self.log_prod,
        }

    def load_protobuf(self, message):
        self.count = int(message.count)
        self.sum = int(message.sum)
        self.log_prod = float(message.log_prod)

    def dump_protobuf(self, message):
        message.count = self.count
        message.sum = self.sum
        message.log_prod = self.log_prod


class Sampler(object):
    def init(self, shared, group=None):
        post = shared if group is None else shared.plus_group(group)
        self.lambda_ = sample_gamma(post.alpha, 1.0 / post.inv_beta)

    def eval(self, shared):
        return sample_poisson(self.lambda_)


def sample_group(shared, size):
    group = Group()
    group.init(shared)
    sampler = Sampler()
    sampler.init(shared, group)
    return [sampler.eval(shared) for _ in xrange(size)]



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import itertools
import numpy
import scipy
from nose.tools import (
    assert_less,
    assert_equal,
    assert_almost_equal,
    assert_raises,
)
from distributions.dbg.random import sample_stick, sample_discrete_log
from distributions.tests.util import (
    require_cython,
    assert_close,
    assert_samples_match_scores,
)


SAMPLES = 1000


def assert_normal(x, y, sigma, stddevs=4.0):
    '''
    Assert that the difference between two values is within a few standard
    deviations of the predicted [normally distributed] error of zero.
    '''
    assert_less(x, y + sigma * stddevs)
    assert_less(y, x + sigma * stddevs)


def test_seed():
    require_cython()
    import distributions.hp.random
    global_rng = distributions.hp.random.random
    distributions.hp.random.seed(0)
    values1 = [global_rng() for _ in xrange(10)]
    distributions.hp.random.seed(0)
    values2 = [global_rng() for _ in xrange(10)]
    assert_equal(values1, values2)


def test_sample_discrete_log_underflow():
    sample_discrete_log([-1e3])
    sample_discrete_log([-1e3, -1e-3])


def test_sample_discrete_log():
    assert_equal(sample_discrete_log([-1.]), 0)
    assert_equal(sample_discrete_log([-1e3]), 0)
    assert_equal(sample_discrete_log([-1e-3]), 0)
    assert_equal(sample_discrete_log([-1., -1e3]), 0)
    assert_equal(sample_discrete_log([-1e3, -1.]), 1)
    assert_raises(Exception, sample_discrete_log, [])


def test_sample_stick():
    gammas = [.1, 1., 5., 10.]
    for gamma in gammas:
        for _ in range(5):
            betas = sample_stick(gamma).values()
            assert_almost_equal(sum(betas), 1., places=5)


def scipy_normal_draw(mean, variance):
    return scipy.stats.norm.rvs(mean, numpy.sqrt(variance))


def test_normal_draw():
    require_cython()
    import distributions.hp.random
    means = [1.0 * i for i in range(-2, 3)]
    variances = [10.0 ** i for i in range(-3, 4)]
    for mean, variance in itertools.product(means, variances):
        # Assume scipy.stats is correct
        #yield _test_normal_draw, scipy_normal_draw, mean, variance
        _test_normal_draw(
            distributions.hp.random.sample_normal,
            mean,
            variance)


def _test_normal_draw(draw, mean, variance):
    samples = [draw(mean, variance) for _ in range(SAMPLES)]
    assert_normal(numpy.mean(samples), mean, numpy.sqrt(variance / SAMPLES))
    error = numpy.array(samples) - mean
    chisq = numpy.dot(error, error) / variance
    assert_normal(chisq, SAMPLES, 2 * SAMPLES)


def test_chisq_draw():
    require_cython()
    import distributions.hp.random
    nus = [1.5 ** i for i in range(-10, 11)]
    for nu in nus:
        # Assume scipy.stats is correct
        #yield _test_chisq_draw, scipy.stats.chi2.rvs, nu
        _test_chisq_draw(distributions.hp.random.sample_chisq, nu)


def _test_chisq_draw(draw, nu):
    samples = [draw(nu) for _ in range(SAMPLES)]
    assert_normal(numpy.mean(samples), nu, numpy.sqrt(2 * nu / SAMPLES))


def test_sample_pair_from_urn():
    require_cython()
    import distributions.lp.random
    TEST_FAIL_PROB = 1e-5
    ITEM_COUNT = 10

    items = range(ITEM_COUNT)
    counts = {(i, j): 0 for i in items for j in items if i != j}
    pair_count = len(counts)

    def test_fail_prob(sample_count):
        '''
        Let X1,...,XK ~iid uniform({1, ..., N = pair_count})
        and for n in {1,..,N} let Cn = sum_k (1 if Xk = n else 0).
        Then for each n,

            P(Cn = 0) = ((N-1) / N)^K
            P(Cn > 0) = 1 - ((N-1) / N)^K
            P(test fails) = 1 - P(for all n, Cn > 0)
                          ~ 1 - (1 - ((N-1) / N)^K)^N
        '''
        item_fail_prob = ((pair_count - 1.0) / pair_count) ** sample_count
        test_fail_prob = 1 - (1 - item_fail_prob) ** pair_count
        return test_fail_prob

    sample_count = 1
    while test_fail_prob(sample_count) > TEST_FAIL_PROB:
        sample_count *= 2
    print 'pair_count = {}'.format(pair_count)
    print 'sample_count = {}'.format(sample_count)

    for _ in xrange(sample_count):
        i, j = distributions.lp.random.sample_pair_from_urn(items)
        assert i != j
        counts[i, j] += 1

    assert_less(0, min(counts.itervalues()))


def test_prob_from_scores():
    require_cython()
    import distributions.lp.random
    rng1 = distributions.lp.random.RNG(0)
    rng2 = distributions.lp.random.RNG(0)
    for size in range(1, 100):
        scores = numpy.random.normal(size=size).tolist()
        for _ in xrange(size):
            sample, prob1 = distributions.lp.random.sample_prob_from_scores(
                rng1,
                scores)
            assert 0 <= sample and sample < size
            prob2 = distributions.lp.random.prob_from_scores(
                rng2,
                sample,
                scores)
            assert_close(
                prob1,
                prob2,
                err_msg='sample_prob_from_scores != prob_from_scores')


def test_sample_prob_from_scores():
    require_cython()
    import distributions.lp.random
    rng = distributions.lp.random.RNG(0)
    for size in range(1, 10):
        scores = numpy.random.normal(size=size).tolist()

        def sampler():
            return distributions.lp.random.sample_prob_from_scores(rng, scores)

        assert_samples_match_scores(sampler)


def test_log_sum_exp():
    require_cython()
    import distributions.lp.random

    for size in xrange(20):
        scores = numpy.random.normal(size=size).tolist()
        expected = numpy.logaddexp.reduce(scores) if size else 0.0
        actual = distributions.lp.random.log_sum_exp(scores)
        assert_close(actual, expected, err_msg='log_sum_exp')



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from scipy.special import betaln, gammaln
from matplotlib import pyplot
from loggamma import gammaln_approx_second_order


def test_betaln(x, y):
    return (gammaln_approx_second_order(x)
            + gammaln_approx_second_order(y)
            - gammaln_approx_second_order(x + y))


def test_pp():
    N = 20

    ms = np.linspace(1, 1e1, N)
    ns = np.linspace(1, 1e1, N)
    alphas = np.linspace(1, 1e1, N)
    betas = np.linspace(1, 1e1, N)

    TERMS = 17

    A = np.zeros((N * N * N * N, TERMS))
    B = np.zeros(N * N * N * N)

    pos = 0
    for mi, m in enumerate(ms):
        print mi
        for ni, n in enumerate(ns):
            for ai, a in enumerate(alphas):
                for bi, b in enumerate(betas):

                    A[pos, :] = (1, m, n, a, b,
                                 m * n, m * a, n * a, m * b, n * b,
                                 np.log(m + a),
                                 np.log(n + b), np.log(m + n + a + b),
                                 np.log(a), np.log(b),
                                 a * np.log(a), b * np.log(b))

                    B[pos] = test_betaln(m + a, n + b) - test_betaln(a, b)

                    pos += 1

    x, residues, rank, s = np.linalg.lstsq(A, B)
    print x
    print residues

    pyplot.plot(((np.dot(A, x) - B) / B) * 100)
    pyplot.ylabel("Percent error")

    pyplot.show()


def test_pp2():
    for MAX in [1e1, 1e2, 1e3, 1e5, 1e7]:
        N = 20
        ms = np.linspace(1, MAX, N)
        ns = np.linspace(1, MAX, N)
        alphas = np.linspace(1, MAX, N)
        betas = np.linspace(1, MAX, N)

        A = np.zeros(N * N * N * N)
        B = np.zeros(N * N * N * N)

        pos = 0
        for mi, m in enumerate(ms):
            print mi
            for ni, n in enumerate(ns):
                for ai, a in enumerate(alphas):
                    for bi, b in enumerate(betas):

                        B[pos] = betaln(m + a, n + b) - betaln(a, b)
                        A[pos] = test_betaln(m + a, n + b) - test_betaln(a, b)

                        pos += 1

        pyplot.figure()
        pyplot.plot(((A - B) / B) * 100)
        pyplot.ylabel("Percent error")
        pyplot.title("MAX=%f" % MAX)

    pyplot.show()


def test2():
    N = 1000

    x = np.linspace(1, 1e1, N)

    y = gammaln(x)
    u = x - 0.5
    yhat = np.log(2 * 3.1415) / 2. + u * np.log(u) - u
    pyplot.plot(x, y)
    pyplot.plot(x, yhat)
    pyplot.show()



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
\cite{teh2006hierarchical}
"""

from itertools import izip
from distributions.dbg.special import log, gammaln
from distributions.dbg.random import (
    sample_discrete,
    sample_dirichlet,
    sample_beta,
)
from distributions.mixins import SharedMixin, GroupIoMixin, SharedIoMixin


NAME = 'DirichletProcessDiscrete'
EXAMPLES = [
    {
        'shared': {
            'gamma': 0.5,
            'alpha': 0.5,
            'betas': {
                0: 0.25,
                7: 0.5,
                8: 0.25,
            },
            'counts': {
                0: 1,
                7: 2,
                8: 4,
            },
        },
        'values': [0, 7, 0, 8, 0, 7, 0],
    },
    {
        'shared': {
            'gamma': 2.0,
            'alpha': 2.0,
            'betas': {},
            'counts': {},
        },
        'values': [5, 4, 3, 2, 1, 0, 3, 2, 1],
    },
]
OTHER = 0xFFFFFFFF
Value = int


class Shared(SharedMixin, SharedIoMixin):
    def __init__(self):
        self.gamma = None
        self.alpha = None
        self.beta0 = None
        self.betas = None
        self.counts = None

    def _load_beta0(self):
        self.beta0 = max(0.0, 1.0 - sum(self.betas.itervalues()))
        if not (self.beta0 <= 1):
            raise ValueError('beta0 out of bounds: {}'.format(self.beta0))
        if self.betas:
            min_beta = min(self.betas.itervalues())
            max_beta = max(self.betas.itervalues())
            if not (0 <= min_beta and max_beta <= 1):
                raise ValueError('betas out of bounds: {}'.format(self.betas))

    def load(self, raw):
        self.gamma = float(raw['gamma'])
        self.alpha = float(raw['alpha'])
        self.betas = {
            int(value): float(beta)
            for value, beta in raw['betas'].iteritems()
        }
        self.counts = {
            int(value): int(count)
            for value, count in raw['counts'].iteritems()
        }
        self._load_beta0()

    def dump(self):
        return {
            'gamma': self.gamma,
            'alpha': self.alpha,
            'betas': self.betas.copy(),
            'counts': self.counts.copy(),
        }

    def load_protobuf(self, message):
        assert len(message.betas) == len(message.values), "invalid message"
        assert len(message.counts) == len(message.values), "invalid message"
        self.gamma = float(message.gamma)
        self.alpha = float(message.alpha)
        self.betas = {
            int(value): float(beta)
            for value, beta in izip(message.values, message.betas)
        }
        self.counts = {
            int(value): int(count)
            for value, count in izip(message.values, message.counts)
        }
        self._load_beta0()

    def dump_protobuf(self, message):
        message.Clear()
        message.gamma = self.gamma
        message.alpha = self.alpha
        for value, beta in self.betas.iteritems():
            message.values.append(value)
            message.betas.append(beta)
            message.counts.append(self.counts[value])

    def add_value(self, value):
        assert value != OTHER, 'cannot add OTHER'
        count = self.counts.get(value, 0) + 1
        self.counts[value] = count
        if count == 1:
            beta = self.beta0 * sample_beta(1.0, self.gamma)
            self.beta0 = max(0.0, self.beta0 - beta)
            self.betas[value] = beta

    def remove_value(self, value):
        assert value != OTHER, 'cannot remove OTHER'
        count = self.counts[value] - 1
        if count:
            self.counts[value] = count
        else:
            del self.counts[value]
            self.beta0 += self.betas.pop(value)

    def realize(self):
        max_size = 10000
        min_beta0 = 1e-4
        new_value = 1 + max(self.betas.iterkeys()) if self.betas else 0
        while len(self.betas) < max_size - 1 and self.beta0 > min_beta0:
            self.add_value(new_value)
            new_value += 1
        if self.beta0 > 0:
            self.add_value(new_value)
            self.betas[new_value] += self.beta0
            self.beta0 = 0


class Group(GroupIoMixin):
    def __init__(self):
        self.counts = None
        self.total = None

    def init(self, shared):
        self.counts = {}  # sparse
        self.total = 0

    def add_repeated_value(self, shared, value, count):
        assert value != OTHER, 'cannot add OTHER'
        assert value in shared.betas, 'unknown value: {}'.format(value)
        if count:
            self.total += count
            try:
                count += self.counts[value]
                if count:
                    self.counts[value] = count
                else:
                    del self.counts[value]
            except KeyError:
                self.counts[value] = count

    def add_value(self, shared, value):
        self.add_repeated_value(shared, value, 1)

    def remove_value(self, shared, value):
        self.add_repeated_value(shared, value, -1)

    def score_value(self, shared, value):
        """
        Adapted from dd.py, which was adapted from:
        McCallum, et. al, 'Rethinking LDA: Why Priors Matter' eqn 4
        """
        denom = shared.alpha + self.total
        if value == OTHER:
            numer = shared.beta0 * shared.alpha
        else:
            count = self.counts.get(value, 0)
            assert count >= 0, "cannot score while in debt"
            numer = shared.betas[value] * shared.alpha + count
        return log(numer / denom)

    def score_data(self, shared):
        assert len(shared.betas), 'betas is empty'
        """
        See doc/dpd.pdf Equation (3)
        """
        score = 0.
        for i, count in self.counts.iteritems():
            assert count >= 0, "cannot score while in debt"
            prior_i = shared.betas[i] * shared.alpha
            score += gammaln(prior_i + count) - gammaln(prior_i)
        score += gammaln(shared.alpha) - gammaln(shared.alpha + self.total)
        return score

    def sample_value(self, shared):
        sampler = Sampler()
        sampler.init(shared, self)
        return sampler.eval(shared)

    def merge(self, shared, source):
        for i, count in source.counts.iteritems():
            self.add_repeated_value(shared, i, count)
        self.total += source.total

    def load(self, raw):
        self.counts = {}
        self.total = 0
        for i, count in raw['counts'].iteritems():
            if count:
                self.counts[int(i)] = int(count)
                self.total += count

    def dump(self):
        counts = {
            value: count
            for value, count in self.counts.iteritems()
            if count
        }
        return {'counts': counts}

    def load_protobuf(self, message):
        self.counts = {}
        self.total = 0
        for i, count in izip(message.keys, message.values):
            if count:
                self.counts[int(i)] = int(count)
                self.total += count

    def dump_protobuf(self, message):
        message.Clear()
        for i, count in self.counts.iteritems():
            if count:
                message.keys.append(i)
                message.values.append(count)


class Sampler(object):
    def init(self, shared, group=None):
        self.values = []
        post = []
        alpha = shared.alpha
        counts = {} if group is None else group.counts
        for value, beta in shared.betas.iteritems():
            self.values.append(value)
            post.append(beta * alpha + counts.get(value, 0))
        if shared.beta0 > 0:
            self.values.append(OTHER)
            post.append(shared.beta0 * alpha)
        self.probs = sample_dirichlet(post)

    def eval(self, shared):
        index = sample_discrete(self.probs)
        return self.values[index]


def sample_group(shared, size):
    group = Group()
    group.init(shared)
    sampler = Sampler()
    sampler.init(shared, group)
    return [sampler.eval(shared) for _ in xrange(size)]



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from libc.math cimport exp
from libcpp.vector cimport vector
from libcpp.utility cimport pair
cimport numpy
numpy.import_array()
from cpython cimport PyObject
from distributions.rng_cc cimport rng_t
from distributions.global_rng cimport get_rng

ctypedef PyObject* O


cdef extern from "distributions/random.hpp" namespace "distributions":
    cdef float log_sum_exp_cc "distributions::log_sum_exp" (
            vector[float] & scores) nogil
    cdef pair[size_t, float] sample_prob_from_scores_overwrite (
            rng_t & rng,
            vector[float] & scores) nogil
    cdef float score_from_scores_overwrite (
            rng_t & rng,
            size_t sample,
            vector[float] & scores) nogil
    cdef pair[O, O] sample_pair_from_urn_cc \
            "distributions::sample_pair_from_urn<PyObject *>" (
            rng_t & rng,
            vector[O] & urn) nogil
    cdef size_t sample_discrete_cc "distributions::sample_discrete" (
            rng_t & rng,
            size_t dim,
            const float * probs) nogil

cdef class RNG:
    cdef rng_t * ptr
    def __cinit__(self):
        self.ptr = new rng_t()
    def __dealloc__(self):
        del self.ptr
    def seed(self, int n):
        self.ptr.seed(n)
    def __call__(self):
        return self.ptr.sample()
    def copy(self):
        cdef RNG result = RNG()
        del result.ptr
        result.ptr = new rng_t(self.ptr[0])
        return result


def log_sum_exp(list scores):
    cdef vector[float] _scores = scores
    return log_sum_exp_cc(_scores)


def sample_prob_from_scores(RNG rng, list scores):
    cdef vector[float] _scores = scores
    return sample_prob_from_scores_overwrite(rng.ptr[0], _scores)


def prob_from_scores(RNG rng, int sample, list scores):
    cdef vector[float] _scores = scores
    cdef float score = score_from_scores_overwrite(rng.ptr[0], sample, _scores)
    cdef float prob = exp(score)
    return prob


def sample_pair_from_urn(list urn):
    cdef vector[O] _urn
    for item in urn:
        _urn.push_back(<O> item)
    cdef pair[O, O] result = sample_pair_from_urn_cc(get_rng()[0], _urn)
    return (<object> result.first, <object> result.second)


def sample_discrete(RNG rng, numpy.ndarray[numpy.float32_t, ndim=1] probs):
    cdef size_t size = probs.shape[0]
    cdef float * data = <float *> probs.data
    return sample_discrete_cc(rng.ptr[0], size, data)



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
import functools
from collections import defaultdict
import numpy
import numpy.random
from nose import SkipTest
from nose.tools import (
    assert_true,
    assert_equal,
    assert_less,
    assert_greater,
    assert_is_instance,
)
from distributions.dbg.random import sample_discrete
from distributions.util import discrete_goodness_of_fit
from distributions.tests.util import (
    require_cython,
    seed_all,
    assert_hasattr,
    assert_close,
)
from distributions.dbg.random import scores_to_probs
import distributions.dbg.clustering
require_cython()
import distributions.lp.clustering
from distributions.lp.clustering import count_assignments
from distributions.lp.mixture import MixtureIdTracker

MODELS = {
    'dbg.LowEntropy': distributions.dbg.clustering.LowEntropy,
    'lp.PitmanYor': distributions.lp.clustering.PitmanYor,
    'lp.LowEntropy': distributions.lp.clustering.LowEntropy,
}

SKIP_EXPENSIVE_TESTS = False
SAMPLE_COUNT = 2000
MIN_GOODNESS_OF_FIT = 1e-3


def iter_examples(Model):
    assert_hasattr(Model, 'EXAMPLES')
    EXAMPLES = Model.EXAMPLES
    assert_is_instance(EXAMPLES, list)
    assert_true(EXAMPLES, 'no examples provided')
    for i, EXAMPLE in enumerate(EXAMPLES):
        print 'example {}/{}'.format(1 + i, len(Model.EXAMPLES))
        yield EXAMPLE


def for_each_model(*filters):
    '''
    Run one test per Model, filtering out inappropriate Models for test.
    '''
    def filtered(test_fun):

        @functools.wraps(test_fun)
        def test_one_model(name):
            Model = MODELS[name]
            for EXAMPLE in iter_examples(Model):
                seed_all(0)
                if SKIP_EXPENSIVE_TESTS and name.startswith('dbg'):
                    sample_count = SAMPLE_COUNT / 10
                else:
                    sample_count = SAMPLE_COUNT
                test_fun(Model, EXAMPLE, sample_count)

        @functools.wraps(test_fun)
        def test_all_models():
            for name, Model in sorted(MODELS.iteritems()):
                if all(f(Model) for f in filters):
                    yield test_one_model, name

        return test_all_models
    return filtered


def canonicalize(assignments):
    groups = defaultdict(lambda: [])
    for value, group in enumerate(assignments):
        groups[group].append(value)
    result = []
    for group in groups.itervalues():
        group.sort()
        result.append(tuple(group))
    result.sort()
    return tuple(result)


@for_each_model()
def test_load_and_dump(Model, EXAMPLE, *unused):
    model = Model()
    model.load(EXAMPLE)
    expected = EXAMPLE
    actual = model.dump()
    assert_close(expected, actual)


def iter_valid_sizes(example, max_size, min_size=2):
    max_size = 5
    dataset_size = example.get('dataset_size', float('inf'))
    sizes = [
        size
        for size in xrange(min_size, max_size + 1)
        if size <= dataset_size
    ]
    assert sizes, 'no valid sizes to test'
    for size in sizes:
        print 'sample_size = {}'.format(size)
        yield size


@for_each_model()
def test_sample_matches_score_counts(Model, EXAMPLE, sample_count):
    for size in iter_valid_sizes(EXAMPLE, max_size=10):
        model = Model()
        model.load(EXAMPLE)

        samples = []
        probs_dict = {}
        for _ in xrange(sample_count):
            value = model.sample_assignments(size)
            sample = canonicalize(value)
            samples.append(sample)
            if sample not in probs_dict:
                assignments = dict(enumerate(value))
                counts = count_assignments(assignments)
                prob = math.exp(model.score_counts(counts))
                probs_dict[sample] = prob

        # renormalize here; test normalization separately
        total = sum(probs_dict.values())
        for key in probs_dict:
            probs_dict[key] /= total

        gof = discrete_goodness_of_fit(samples, probs_dict, plot=True)
        print '{} gof = {:0.3g}'.format(Model.__name__, gof)
        assert_greater(gof, MIN_GOODNESS_OF_FIT)


@for_each_model()
def test_score_counts_is_normalized(Model, EXAMPLE, sample_count):

    for sample_size in iter_valid_sizes(EXAMPLE, max_size=10):
        model = Model()
        model.load(EXAMPLE)

        if Model.__name__ == 'LowEntropy' and sample_size < model.dataset_size:
            print 'WARNING LowEntropy.score_counts normalization is imprecise'
            print '  when sample_size < dataset_size'
            tol = 0.5
        else:
            tol = 0.01

        probs_dict = {}
        for _ in xrange(sample_count):
            value = model.sample_assignments(sample_size)
            sample = canonicalize(value)
            if sample not in probs_dict:
                assignments = dict(enumerate(value))
                counts = count_assignments(assignments)
                prob = math.exp(model.score_counts(counts))
                probs_dict[sample] = prob

        total = sum(probs_dict.values())
        assert_less(abs(total - 1), tol, 'not normalized: {}'.format(total))


def add_to_counts(counts, pos):
    counts = counts[:]
    counts[pos] += 1
    return counts


@for_each_model()
def test_score_add_value_matches_score_counts(Model, EXAMPLE, sample_count):
    for sample_size in iter_valid_sizes(EXAMPLE, min_size=2, max_size=10):
        model = Model()
        model.load(EXAMPLE)

        samples = set(
            canonicalize(model.sample_assignments(sample_size - 1))
            for _ in xrange(sample_count)
        )

        for sample in samples:
            nonempty_group_count = len(sample)
            counts = map(len, sample)
            actual = numpy.zeros(len(counts) + 1)
            expected = numpy.zeros(len(counts) + 1)

            # add to existing group
            for i, group in enumerate(sample):
                group_size = len(sample[i])
                expected[i] = model.score_counts(add_to_counts(counts, i))
                actual[i] = model.score_add_value(
                    group_size,
                    nonempty_group_count,
                    sample_size - 1)

            # add to new group
            i = len(counts)
            group_size = 0
            expected[i] = model.score_counts(counts + [1])
            actual[i] = model.score_add_value(
                group_size,
                nonempty_group_count,
                sample_size - 1)

            actual = scores_to_probs(actual)
            expected = scores_to_probs(expected)
            print actual, expected
            assert_close(actual, expected, tol=0.05)


@for_each_model(lambda Model: hasattr(Model, 'Mixture'))
def test_mixture_score_matches_score_add_value(Model, EXAMPLE, *unused):
    sample_count = 200
    model = Model()
    model.load(EXAMPLE)

    if Model.__name__ == 'LowEntropy' and sample_count > model.dataset_size:
        raise SkipTest('skipping trivial example')

    assignment_vector = model.sample_assignments(sample_count)
    assignments = dict(enumerate(assignment_vector))
    nonempty_counts = count_assignments(assignments)
    nonempty_group_count = len(nonempty_counts)
    assert_greater(nonempty_group_count, 1, "test is inaccurate")

    def check_counts(mixture, counts, empty_group_count):
        #print 'counts =', counts
        empty_groupids = frozenset(mixture.empty_groupids)
        assert_equal(len(empty_groupids), empty_group_count)
        for groupid in empty_groupids:
            assert_equal(counts[groupid], 0)

    def check_scores(mixture, counts, empty_group_count):
        sample_count = sum(counts)
        nonempty_group_count = len(counts) - empty_group_count
        expected = [
            model.score_add_value(
                group_size,
                nonempty_group_count,
                sample_count,
                empty_group_count)
            for group_size in counts
        ]
        noise = numpy.random.randn(len(counts))
        actual = numpy.zeros(len(counts), dtype=numpy.float32)
        actual[:] = noise
        mixture.score_value(model, actual)
        assert_close(actual, expected)
        return actual

    for empty_group_count in [1, 10]:
        print 'empty_group_count =', empty_group_count
        counts = nonempty_counts + [0] * empty_group_count
        numpy.random.shuffle(counts)
        mixture = Model.Mixture()
        id_tracker = MixtureIdTracker()

        print 'init'
        mixture.init(model, counts)
        id_tracker.init(len(counts))
        check_counts(mixture, counts, empty_group_count)
        check_scores(mixture, counts, empty_group_count)

        print 'adding'
        groupids = []
        for _ in xrange(sample_count):
            check_counts(mixture, counts, empty_group_count)
            scores = check_scores(mixture, counts, empty_group_count)
            probs = scores_to_probs(scores)
            groupid = sample_discrete(probs)
            expected_group_added = (counts[groupid] == 0)
            counts[groupid] += 1
            actual_group_added = mixture.add_value(model, groupid)
            assert_equal(actual_group_added, expected_group_added)
            groupids.append(groupid)
            if actual_group_added:
                id_tracker.add_group()
                counts.append(0)

        check_counts(mixture, counts, empty_group_count)
        check_scores(mixture, counts, empty_group_count)

        print 'removing'
        for global_groupid in groupids:
            groupid = id_tracker.global_to_packed(global_groupid)
            counts[groupid] -= 1
            expected_group_removed = (counts[groupid] == 0)
            actual_group_removed = mixture.remove_value(model, groupid)
            assert_equal(actual_group_removed, expected_group_removed)
            if expected_group_removed:
                id_tracker.remove_group(groupid)
                back = counts.pop()
                if groupid < len(counts):
                    counts[groupid] = back
            check_counts(mixture, counts, empty_group_count)
            check_scores(mixture, counts, empty_group_count)



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from libc.stdint cimport uint32_t


cdef extern from "distributions/mixture.hpp":
    cppclass MixtureIdTracker_cc "distributions::MixtureIdTracker":
        void init (size_t group_count) nogil except +
        void add_group () nogil except +
        void remove_group (uint32_t packed) nogil except +
        uint32_t packed_to_global (uint32_t packed) nogil except +
        uint32_t global_to_packed (uint32_t packed) nogil except +


cdef class MixtureIdTracker:
    cdef MixtureIdTracker_cc * ptr
    def __cinit__(self):
        self.ptr = new MixtureIdTracker_cc()
    def __dealloc__(self):
        del self.ptr

    def init(self, int group_count=0):
        self.ptr.init(group_count)

    def add_group(self):
        self.ptr.add_group()

    def remove_group(self, int packed):
        self.ptr.remove_group(packed)

    def packed_to_global(self, int packed):
        return self.ptr.packed_to_global(packed)

    def global_to_packed(self, int global_):
        return self.ptr.global_to_packed(global_)



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

cimport _dpd
import _dpd

from cython.operator cimport dereference as deref, preincrement as inc
from distributions.sparse_counter cimport SparseCounter, SparseFloat
from distributions.mixins import SharedMixin, GroupIoMixin, SharedIoMixin


NAME = 'DirichletProcessDiscrete'
EXAMPLES = [
    {
        'shared': {
            'gamma': 0.5,
            'alpha': 0.5,
            'betas': {
                0: 0.25,
                7: 0.5,
                8: 0.25,
            },
            'counts': {
                0: 1,
                7: 2,
                8: 4,
            },
        },
        'values': [0, 7, 0, 8, 0, 7, 0],
    },
    {
        'shared': {
            'gamma': 2.0,
            'alpha': 2.0,
            'betas': {},
            'counts': {},
        },
        'values': [5, 4, 3, 2, 1, 0, 3, 2, 1],
    },
]
Value = int


cdef class _Shared(_dpd.Shared):
    def load(self, dict raw):
        self.ptr.gamma = raw['gamma']
        self.ptr.alpha = raw['alpha']
        self.ptr.betas.clear()
        self.ptr.counts.clear()
        cdef dict raw_betas = raw['betas']
        cdef dict raw_counts = raw['counts']
        cdef int value
        cdef float beta
        cdef double beta0 = 1.0
        for value, beta in raw_betas.iteritems():
            self.ptr.betas.add(int(value), beta)
            beta0 -= beta
        self.ptr.beta0 = beta0
        cdef int count
        for value, count in raw_counts.iteritems():
            self.ptr.counts.add(int(value), count)

    def dump(self):
        cdef dict betas = {}
        cdef dict counts = {}
        cdef SparseFloat.iterator it = self.ptr.betas.begin()
        cdef SparseFloat.iterator end = self.ptr.betas.end()
        cdef int value
        while it != end:
            value = deref(it).first
            betas[value] = float(deref(it).second)
            counts[value] = int(self.ptr.counts.get_count(value))
            inc(it)
        return {
            'gamma': float(self.ptr.gamma),
            'alpha': float(self.ptr.alpha),
            'betas': betas,
            'counts': counts,
        }

    def load_protobuf(self, message):
        self.ptr.gamma = message.gamma
        self.ptr.alpha = message.alpha
        self.ptr.betas.clear()
        self.ptr.counts.clear()
        cdef int i
        cdef int value
        cdef float beta
        cdef double beta0 = 1.0
        for i in xrange(len(message.betas)):
            value = message.values[i]
            beta = message.betas[i]
            self.ptr.betas.add(value, beta)
            self.ptr.counts.add(value, message.counts[i])
            beta0 -= beta
        self.ptr.beta0 = beta0

    def dump_protobuf(self, message):
        message.Clear()
        message.gamma = self.ptr.gamma
        message.alpha = self.ptr.alpha
        cdef SparseFloat.iterator it = self.ptr.betas.begin()
        cdef SparseFloat.iterator end = self.ptr.betas.end()
        cdef int value
        while it != end:
            value = deref(it).first
            message.values.append(value)
            message.betas.append(deref(it).second)
            message.counts.append(self.ptr.counts.get_count(value))
            inc(it)


#class Shared(_Shared, SharedMixin, SharedIoMixin):
class Shared(_Shared, SharedIoMixin):
    pass


cdef class _Group(_dpd.Group):
    def load(self, dict raw):
        cdef SparseCounter * counts = & self.ptr.counts
        counts.clear()
        cdef dict raw_counts = raw['counts']
        cdef int value
        cdef int count
        for value, count in raw_counts.iteritems():
            counts.init_count(value, count)

    def dump(self):
        cdef dict counts = {}
        cdef SparseCounter.iterator it = self.ptr.counts.begin()
        cdef SparseCounter.iterator end = self.ptr.counts.end()
        while it != end:
            counts[int(deref(it).first)] = deref(it).second
            inc(it)
        return {'counts': counts}


class Group(_Group, GroupIoMixin):
    pass


class Sampler(_dpd.Sampler):
    pass


Mixture = _dpd.Mixture
sample_group = _dpd.sample_group



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from distributions.dbg.special import log
from distributions.dbg.random import sample_discrete_log
from distributions.mixins import SharedIoMixin


class LowEntropy(SharedIoMixin):
    '''
    A parameter-free clustering prior based on partition entropy.

    See /derivations/clustering.py for intro and motivation.

    Definition
    ----------

    Let X = [X_0,...,X_{N-1}] be an assignment vector of N points into
    clusters of sizes [N_0,...,N_{K-1}], i.e.

        N_k = #{i | X_i = k, i in {0,...,N-1}}

    The empirical entropy of the assignment vector is

        H(X) = sum_k -p_k log p_k  = sum_k N_k/N log( N_k/N )

    And so the information complexity of the assignment vector is N H(X),
    where each assignment requires H(X) information to specify.
    Define the prior probability in terms of cluster sizes [N_1,...,N_k] as

        P(X)   propto   exp(-N H(X))   propto   prod_{k=1}^K N_k^{N_k}

    Implementation
    --------------

    This class implements approximations to numerical functions
    required in MCMC inference using this clustering prior:

    - evaluating log posterior predictive probability of an assignment
      given a partial assignment vector

        score_add_value(...) =

            log P[ X_i=x_i | X_0=x_0,...,X_{i-1}=x_{i-1} ] Z / Z_approx

    - evaluating log probability of a full assignment vector,
      up to the constant partition function

        score_counts(...) =

            log P[ X_0=x_0,...,X_{n-1}=x_{n-1} ] Z(n) / Z_approx(n)

    - sampling a full assignment vector X ~ P[ X=x ],

        X = model.sample_assignments(...)

    - evaluating the partition function Z(n) exactly for small n
      and at low accuracy for large n

        log_partition_function(n) = log Z_exact(n)  for n <= 47
                                    log Z_approx(n) for n > 47
    '''

    def __init__(self, dataset_size=0):
        self.dataset_size = int(dataset_size)

    #-------------------------------------------------------------------------
    # Serialization

    def load(self, raw):
        self.dataset_size = int(raw['dataset_size'])
        assert self.dataset_size >= 0

    def dump(self):
        return {'dataset_size': self.dataset_size}

    def load_protobuf(self, message):
        self.dataset_size = int(message.dataset_size)

    def dump_protobuf(self, message):
        message.Clear()
        message.dataset_size = self.dataset_size

    #-------------------------------------------------------------------------
    # Sampling

    def sample_assignments(self, sample_size):
        '''
        Sample partial assignment vector

            [X_0, ..., X_{n-1}]

        where

            n = sample_size <= N = dataset_size.
        '''
        assert sample_size <= self.dataset_size

        assignments = []
        counts = []
        scores = []
        bogus = 0

        for size in xrange(sample_size):

            score_empty = self.score_add_value(0, bogus, size)
            if len(counts) == 0 or counts[-1] != 0:
                counts.append(0)
                scores.append(score_empty)
            else:
                scores[-1] = score_empty

            assign = sample_discrete_log(scores)
            counts[assign] += 1
            size += 1
            scores[assign] = self.score_add_value(counts[assign], bogus, bogus)
            assignments.append(assign)

        return assignments

    #-------------------------------------------------------------------------
    # Scoring

    def score_counts(self, counts):
        '''
        Return log probability of data, given sufficient statistics of
        a partial assignment vector [X_0,...,X_{n-1}]

            log P[ X_0=x_0, ..., X_{n-1}=x_{n-1} ]
        '''
        score = 0.0
        sample_size = 0
        for count in counts:
            sample_size += count
            if count > 1:
                score += count * log(count)
        assert sample_size <= self.dataset_size

        if sample_size != self.dataset_size:
            log_factor = self._approximate_postpred_correction(sample_size)
            score += log_factor * (len(counts) - 1)
            score += self._approximate_dataprob_correction(sample_size)
        score -= self.log_partition_function(sample_size)
        return score

    def score_add_value(
            self,
            group_size,
            nonempty_group_count,
            sample_size,
            empty_group_count=1):
        '''
        Return log of posterior predictive probability given
        sufficient statistics of a partial assignments vector [X_0,...,X_{n-1}]

            log P[ X_n = k | X_0=x_0, ..., X_{n-1}=x_{n-1} ]

        where

            group_size = #{i | x_i = k, i in {0,...,n-1}}

            nonempty_group_count = #{x_i | i in {0,...,n-1}}

            sample_size = n

        and empty_group_count is the number of empty groups that are uniformly
        competing for the assignment.  Typically empty_group_count = 1, but
        multiple empty "ephemeral" groups are used in e.g. Radford Neal's
        Algorithm-8 \cite{neal2000markov}.
        '''
        assert sample_size < self.dataset_size
        assert 0 < empty_group_count

        if group_size == 0:
            score = -log(empty_group_count)
            if sample_size + 1 < self.dataset_size:
                score += self._approximate_postpred_correction(sample_size + 1)
            return score

        # see `python derivations/clustering.py fastlog`
        very_large = 10000
        bigger = 1.0 + group_size
        if group_size > very_large:
            return 1.0 + log(bigger)
        else:
            return log(bigger / group_size) * group_size + log(bigger)

    def score_remove_value(
            self,
            group_size,
            nonempty_group_count,
            sample_size,
            empty_group_count=1):
        '''
        Reverse transition probability of score_add_value, given
        sufficient statistics of the partial assignment vector  [X_0,...,X_n}]

            -log P[ X_n=x_n | X_0=x_0, ..., X_{n-1}=x_{n-1} ]

        This is useful in Metropolis-Hastings inference.
        '''
        assert sample_size > 0

        group_size -= 1
        return -self.score_add_value(
            group_size,
            nonempty_group_count,
            sample_size,
            empty_group_count)

    #-------------------------------------------------------------------------
    # Approximations

    # this code was generated by derivations/clustering.py
    def log_partition_function(self, sample_size):
        '''
        Computes

            log_sum_exp(
                sum(n * log(n) for n in partition)
                for partition in partitions(sample_size)
            )

        exactly for small n, and approximately for large n.
        '''
        # TODO incorporate dataset_size for higher accuracy
        n = sample_size
        if n < 48:
            return LowEntropy.log_partition_function_table[n]
        else:
            coeff = 0.28269584
            log_z_max = n * log(n)
            return log_z_max * (1.0 + coeff * n ** -0.75)

    # this code was generated by derivations/clustering.py
    log_partition_function_table = [
        0.00000000, 0.00000000, 1.60943791, 3.68887945, 6.07993320,
        8.70549682, 11.51947398, 14.49108422, 17.59827611, 20.82445752,
        24.15668300, 27.58456586, 31.09958507, 34.69462231, 38.36364086,
        42.10145572, 45.90356476, 49.76602176, 53.68533918, 57.65841234,
        61.68245958, 65.75497413, 69.87368527, 74.03652635, 78.24160846,
        82.48719834, 86.77169993, 91.09363859, 95.45164780, 99.84445762,
        104.27088480, 108.72982416, 113.22024112, 117.74116515, 122.29168392,
        126.87093829, 131.47811772, 136.11245629, 140.77322911, 145.45974907,
        150.17136399, 154.90745399, 159.66742919, 164.45072752, 169.25681285,
        174.08517319, 178.93531914, 183.80678238
    ]

    def _approximate_postpred_correction(self, sample_size):
        '''
        ad hoc approximation,
        see `python derivations/clustering.py postpred`
        see `python derivations/clustering.py approximations`
        '''
        assert 0 < sample_size
        assert sample_size < self.dataset_size

        exponent = 0.45 - 0.1 / sample_size - 0.1 / self.dataset_size
        scale = self.dataset_size / sample_size
        return log(scale) * exponent

    def _approximate_dataprob_correction(self, sample_size):
        '''
        ad hoc approximation,
        see `python derivations/clustering.py dataprob`
        see `python derivations/clustering.py approximations`
        '''
        n = log(sample_size)
        N = log(self.dataset_size)
        return 0.061 * n * (n - N) * (n + N) ** 0.75

    #-------------------------------------------------------------------------
    # Examples

    EXAMPLES = [
        {'dataset_size': 5},
        {'dataset_size': 1000},
    ]



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
A conjugate self on normally-distributied univariate data in which the
prior on the mean is normally distributed, and the prior on the variance
is Inverse-Chi-Square distributed.

The equations used here are from \cite{murphy2007conjugate}
Murphy, K. "Conjugate Bayesian analysis of the Gaussian distribution" (2007)
Equation numbers referenced below are from this paper.
"""

import numpy
cimport numpy
numpy.import_array()
from distributions.hp.special cimport sqrt, log, gammaln, M_PI
from distributions.hp.random cimport sample_normal, sample_chisq
from distributions.mixins import SharedMixin, GroupIoMixin, SharedIoMixin


# FIXME how does this relate to distributions.dbg.random.score_student_t
cdef double score_student_t(double x, double nu, double mu, double sigmasq):
    """
    \cite{murphy2007conjugate}, Eq. 304
    """
    cdef double c = gammaln(.5 * (nu + 1.))\
        - (gammaln(.5 * nu) + .5 * (log(nu * M_PI * sigmasq)))
    cdef double xt = (x - mu)
    cdef double s = xt * xt / sigmasq
    cdef double d = -(.5 * (nu + 1.)) * log(1. + s / nu)
    return c + d


NAME = 'NormalInverseChiSq'
EXAMPLES = [
    {
        'shared': {'mu': 0., 'kappa': 1., 'sigmasq': 1., 'nu': 1.},
        'values': [-4.0, -2.0, -1.0, -0.5, 0.0, 0.5, 1.0, 2.0, 4.0],
    },
]
Value = float


ctypedef double _Value


cdef class _Shared:
    cdef double mu
    cdef double kappa
    cdef double sigmasq
    cdef double nu

    cdef _Shared plus_group(self, _Group group):
        """
        \cite{murphy2007conjugate}, Eqs.141-144
        """
        cdef double mu_1 = self.mu - group.mean
        cdef double kappa_n = self.kappa + group.count
        cdef double mu_n = (
            self.kappa * self.mu + group.mean * group.count) / kappa_n
        cdef double nu_n = self.nu + group.count
        cdef double sigmasq_n = 1. / nu_n * (
            self.nu * self.sigmasq
            + group.count_times_variance
            + (group.count * self.kappa * mu_1 * mu_1) / kappa_n)

        cdef _Shared post = _Shared()
        post.mu = mu_n
        post.kappa = kappa_n
        post.nu = nu_n
        post.sigmasq = sigmasq_n
        return post

    def load(self, dict raw):
        self.mu = raw['mu']
        self.kappa = raw['kappa']
        self.sigmasq = raw['sigmasq']
        self.nu = raw['nu']

    def dump(self):
        return {
            'mu': self.mu,
            'kappa': self.kappa,
            'sigmasq': self.sigmasq,
            'nu': self.nu,
        }


class Shared(_Shared, SharedMixin, SharedIoMixin):
    pass


cdef class _Group:
    cdef size_t count
    cdef double mean
    cdef double count_times_variance

    def init(self, _Shared shared):
        self.count = 0
        self.mean = 0.
        self.count_times_variance = 0.

    def add_value(self, _Shared shared, double value):
        self.count += 1
        cdef double delta = value - self.mean
        self.mean += delta / self.count
        self.count_times_variance += delta * (value - self.mean)

    def add_repeated_value(self, _Shared shared, double value, int count):
        self.count += count
        cdef double delta = count * value - self.mean
        self.mean += delta / self.count
        self.count_times_variance += delta * (value - self.mean)

    def remove_value(self, _Shared shared, double value):
        cdef double total = self.mean * self.count
        cdef double delta = value - self.mean
        self.count -= 1
        if self.count == 0:
            self.mean = 0.
        else:
            self.mean = (total - value) / self.count
        if self.count <= 1:
            self.count_times_variance = 0.
        else:
            self.count_times_variance -= delta * (value - self.mean)

    def merge(self, _Shared shared, _Group source):
        cdef size_t count = self.count + source.count
        cdef double delta = source.mean - self.mean
        cdef double source_part = <double> source.count / count
        cdef double cross_part = self.count * source_part
        self.count = count
        self.mean += source_part * delta
        self.count_times_variance += \
            source.count_times_variance + cross_part * delta * delta

    def score_value(self, _Shared shared, _Value value):
        """
        \cite{murphy2007conjugate}, Eq. 176
        """
        cdef _Shared post = shared.plus_group(self)
        return score_student_t(
            value,
            post.nu,
            post.mu,
            ((1 + post.kappa) * post.sigmasq) / post.kappa)

    def score_data(self, _Shared shared):
        """
        \cite{murphy2007conjugate}, Eq. 171
        """
        cdef _Shared post = shared.plus_group(self)
        return gammaln(post.nu / 2.) - gammaln(shared.nu / 2.) + \
            0.5 * log(shared.kappa / post.kappa) + \
            (0.5 * shared.nu) * log(shared.nu * shared.sigmasq) - \
            (0.5 * post.nu) * log(post.nu * post.sigmasq) - \
            self.count / 2. * 1.1447298858493991

    def sample_value(self, _Shared shared):
        cdef Sampler sampler = Sampler()
        sampler.init(shared, self)
        return sampler.eval(shared)

    def load(self, dict raw):
        self.count = raw['count']
        self.mean = raw['mean']
        self.count_times_variance = raw['count_times_variance']

    def dump(self):
        return {
            'count': self.count,
            'mean': self.mean,
            'count_times_variance': self.count_times_variance,
        }


class Group(_Group, GroupIoMixin):
    pass


cdef class Sampler:
    cdef double mu
    cdef double sigmasq

    def init(self, _Shared shared, _Group group=None):
        """
        Draw samples from the marginal posteriors of mu and sigmasq

        \cite{murphy2007conjugate}, Eqs. 156 & 167
        """
        cdef _Shared post
        post = shared if group is None else shared.plus_group(group)
        self.sigmasq = post.nu * post.sigmasq / sample_chisq(post.nu)
        self.mu = sample_normal(post.mu, self.sigmasq / post.kappa)

    def eval(self, _Shared shared):
        return sample_normal(self.mu, self.sigmasq)


def sample_group(_Shared shared, int size):
    cdef Sampler sampler = Sampler()
    sampler.init(shared)
    cdef list result = []
    cdef int i
    for i in xrange(size):
        result.append(sampler.eval(shared))
    return result



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from libcpp.vector cimport vector
from libcpp.utility cimport pair
cimport numpy
numpy.import_array()
from cython import address
from cython.operator cimport dereference as deref, preincrement as inc
from distributions.rng_cc cimport rng_t
from distributions.global_rng cimport get_rng
from distributions.lp.vector cimport VectorFloat, vector_float_to_ndarray
from distributions.mixins import SharedIoMixin


cdef extern from 'distributions/mixture.hpp':
    cppclass IdSet \
            "std::unordered_set<size_t, distributions::TrivialHash<size_t>>":
        cppclass iterator "const_iterator":
            size_t & operator*()
            iterator operator++() nogil
            bint operator!=(iterator) nogil
        iterator begin() nogil
        iterator end() nogil


cdef extern from 'distributions/clustering.hpp':
    cppclass Assignments "distributions::Clustering<int>::Assignments":
        Assignments() nogil except +
        cppclass iterator:
            pair[int, int] & operator*() nogil
            iterator operator++() nogil
            bint operator!=(iterator) nogil
        int & operator[](int) nogil
        iterator begin() nogil
        iterator end() nogil

    cdef vector[int] count_assignments_cc \
            "distributions::Clustering<int>::count_assignments" \
            (Assignments & assignments) nogil except +

    cppclass PitmanYor_cc "distributions::Clustering<int>::PitmanYor":
        float alpha
        float d
        vector[int] sample_assignments(int size, rng_t & rng) nogil except +
        cppclass Mixture:
            size_t size "counts().size" () nogil except +
            IdSet.iterator empty_groupids_begin \
                    "empty_groupids().begin" () nogil except +
            IdSet.iterator empty_groupids_end \
                    "empty_groupids().end" () nogil except +
            void set_counts "counts() = " (vector[int] &) nogil except +
            void init (PitmanYor_cc &) nogil except +
            bint add_value (PitmanYor_cc &, size_t) nogil except +
            bint remove_value (PitmanYor_cc &, size_t) nogil except +
            void score_value (PitmanYor_cc &, VectorFloat &) nogil except +
        float score_counts(vector[int] & counts) nogil except +
        float score_add_value (
                int group_size,
                int nonempty_group_count,
                int sample_size,
                int empty_group_count) nogil except +
        float score_remove_value (
                int group_size,
                int nonempty_group_count,
                int sample_size,
                int empty_group_count) nogil except +

    cppclass LowEntropy_cc "distributions::Clustering<int>::LowEntropy":
        int dataset_size
        vector[int] sample_assignments(int size, rng_t & rng) nogil except +
        cppclass Mixture:
            size_t size "counts().size" () nogil except +
            IdSet.iterator empty_groupids_begin \
                    "empty_groupids().begin" () nogil except +
            IdSet.iterator empty_groupids_end \
                    "empty_groupids().end" () nogil except +
            void set_counts "counts() = " (vector[int] &) nogil except +
            void init (LowEntropy_cc &) nogil except +
            bint add_value (LowEntropy_cc &, size_t) nogil except +
            bint remove_value (LowEntropy_cc &, size_t) nogil except +
            void score_value (LowEntropy_cc &, VectorFloat &) nogil except +
        float score_counts(vector[int] & counts) nogil except +
        float score_add_value (
                int group_size,
                int nonempty_group_count,
                int sample_size,
                int empty_group_count) nogil except +
        float score_remove_value (
                int group_size,
                int nonempty_group_count,
                int sample_size,
                int empty_group_count) nogil except +


cpdef list count_assignments(dict assignments):
    cdef Assignments assignments_cc
    cdef int value_id
    cdef int group_id
    for value_id, group_id in assignments.iteritems():
        assignments_cc[value_id] = group_id
    cdef list counts = count_assignments_cc(assignments_cc)
    return counts


cdef dict dump_assignments(Assignments & assignments):
    cdef dict raw = {}
    cdef Assignments.iterator i = assignments.begin()
    while i != assignments.end():
        assignments[deref(i).first] = deref(i).second
        inc(i)
    return raw


#-----------------------------------------------------------------------------
# Pitman-Yor

cdef class PitmanYor_cy:
    cdef PitmanYor_cc * ptr
    def __cinit__(self):
        self.ptr = new PitmanYor_cc()
    def __dealloc__(self):
        del self.ptr

    def __init__(self, **kwargs):
        if kwargs:
            self.load(kwargs)
        else:
            self.ptr.alpha = 1.0
            self.ptr.d = 0.0

    property alpha:
        def __get__(self):
            return self.ptr.alpha

    property d:
        def __get__(self):
            return self.ptr.d

    def load(self, dict raw):
        cdef float alpha = raw['alpha']
        cdef float d = raw['d']
        assert 0 < alpha
        assert 0 <= d and d < 1
        self.ptr.alpha = alpha
        self.ptr.d = d

    def dump(self):
        return {
            'alpha': self.ptr.alpha,
            'd': self.ptr.d,
        }

    def sample_assignments(self, int size):
        cdef list assignments = self.ptr.sample_assignments(size, get_rng()[0])
        return assignments

    def score_counts(self, list counts):
        cdef vector[int] counts_cc = counts
        cdef float score = self.ptr.score_counts(counts_cc)
        return score

    def score_add_value(
            self,
            int group_size,
            int nonempty_group_count,
            int sample_size,
            int empty_group_count=1):
        return self.ptr.score_add_value(
            group_size,
            nonempty_group_count,
            sample_size,
            empty_group_count)

    def score_remove_value(
            self,
            int group_size,
            int nonempty_group_count,
            int sample_size,
            int empty_group_count=1):
        return self.ptr.score_remove_value(
            group_size,
            nonempty_group_count,
            sample_size,
            empty_group_count)

    EXAMPLES = [
        {'alpha': 1., 'd': 0.},
        {'alpha': 1., 'd': 0.1},
        {'alpha': 1., 'd': 0.9},
        {'alpha': 10., 'd': 0.1},
        {'alpha': 0.1, 'd': 0.1},
    ]


cdef class PitmanYorMixture:
    cdef PitmanYor_cc.Mixture * ptr
    def __cinit__(self):
        self.ptr = new PitmanYor_cc.Mixture()
    def __dealloc__(self):
        del self.ptr

    def __len__(self):
        return self.ptr.size()

    property empty_groupids:
        def __get__(self):
            cdef PitmanYor_cc.Mixture * ptr = self.ptr
            cdef IdSet.iterator i = ptr.empty_groupids_begin()
            cdef IdSet.iterator end = ptr.empty_groupids_end()
            while i != end:
                yield deref(i)
                inc(i)

    def init(self, PitmanYor_cy model, list counts):
        cdef vector[int] counts_cc = counts
        self.ptr.set_counts(counts_cc)
        self.ptr.init(model.ptr[0])

    def add_value(self, PitmanYor_cy model, int groupid):
        return self.ptr.add_value(model.ptr[0], groupid)

    def remove_value(self, PitmanYor_cy model, int groupid):
        return self.ptr.remove_value(model.ptr[0], groupid)

    def score_value(
            self,
            PitmanYor_cy model,
            numpy.ndarray[numpy.float32_t, ndim=1] scores):
        cdef VectorFloat scores_cc
        scores_cc.resize(self.ptr.size())
        self.ptr.score_value(model.ptr[0], scores_cc)
        vector_float_to_ndarray(scores_cc, scores)


class PitmanYor(PitmanYor_cy, SharedIoMixin):

    def load_protobuf(self, message):
        self.load({'alpha': message.alpha, 'd': message.d})

    def dump_protobuf(self, message):
        dumped = self.dump()
        message.Clear()
        message.alpha = dumped['alpha']
        message.d = dumped['d']

    Mixture = PitmanYorMixture


#-----------------------------------------------------------------------------
# Low Entropy

cdef class LowEntropy_cy:
    cdef LowEntropy_cc * ptr
    def __cinit__(self):
        self.ptr = new LowEntropy_cc()
    def __dealloc__(self):
        del self.ptr

    def __init__(self, **kwargs):
        if kwargs:
            self.load(kwargs)
        else:
            self.ptr.dataset_size = 0

    property dataset_size:
        def __get__(self):
            return self.ptr.dataset_size

    def load(self, dict raw):
        cdef int dataset_size = raw['dataset_size']
        assert dataset_size >= 0
        self.ptr.dataset_size = dataset_size

    def dump(self):
        return {'dataset_size': self.ptr.dataset_size}

    def sample_assignments(self, int size):
        cdef list assignments = self.ptr.sample_assignments(size, get_rng()[0])
        return assignments

    def score_counts(self, list counts):
        cdef vector[int] counts_cc = counts
        cdef float score = self.ptr.score_counts(counts_cc)
        return score

    def score_add_value(
            self,
            int group_size,
            int nonempty_group_count,
            int sample_size,
            int empty_group_count=1):
        return self.ptr.score_add_value(
            group_size,
            nonempty_group_count,
            sample_size,
            empty_group_count)

    def score_remove_value(
            self,
            int group_size,
            int nonempty_group_count,
            int sample_size,
            int empty_group_count=1):
        return self.ptr.score_remove_value(
            group_size,
            nonempty_group_count,
            sample_size,
            empty_group_count)

    EXAMPLES = [
        {'dataset_size': 5},
        {'dataset_size': 10},
        {'dataset_size': 100},
        {'dataset_size': 1000},
    ]


cdef class LowEntropyMixture:
    cdef LowEntropy_cc.Mixture * ptr
    def __cinit__(self):
        self.ptr = new LowEntropy_cc.Mixture()
    def __dealloc__(self):
        del self.ptr

    def __len__(self):
        return self.ptr.size()

    property empty_groupids:
        def __get__(self):
            cdef LowEntropy_cc.Mixture * ptr = self.ptr
            cdef IdSet.iterator i = ptr.empty_groupids_begin()
            cdef IdSet.iterator end = ptr.empty_groupids_end()
            while i != end:
                yield deref(i)
                inc(i)

    def init(self, LowEntropy_cy model, list counts):
        cdef vector[int] counts_cc = counts
        self.ptr.set_counts(counts_cc)
        self.ptr.init(model.ptr[0])

    def add_value(self, LowEntropy_cy model, int groupid):
        return self.ptr.add_value(model.ptr[0], groupid)

    def remove_value(self, LowEntropy_cy model, int groupid):
        return self.ptr.remove_value(model.ptr[0], groupid)

    def score_value(
            self,
            LowEntropy_cy model,
            numpy.ndarray[numpy.float32_t, ndim=1] scores):
        cdef VectorFloat scores_cc
        scores_cc.resize(self.ptr.size())
        self.ptr.score_value(model.ptr[0], scores_cc)
        vector_float_to_ndarray(scores_cc, scores)


class LowEntropy(LowEntropy_cy, SharedIoMixin):

    def load_protobuf(self, message):
        self.load({'dataset_size': message.dataset_size})

    def dump_protobuf(self, message):
        dumped = self.dump()
        message.Clear()
        message.datset_size = dumped['dataset_size']

    Mixture = LowEntropyMixture



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from distributions.tests.util import (
    assert_all_close,
    list_models,
    import_model,
)


MODULES = {}
for spec in list_models():
    MODULES.setdefault(spec['name'], []).append(import_model(spec))


def test_model():
    for name in MODULES:
        yield _test_model, name


def _test_model(name):
    modules = MODULES[name]
    assert_all_close([m.NAME for m in modules], err_msg='Model.__name__')
    EXAMPLES = [e for m in modules for e in m.EXAMPLES]
    for EXAMPLE in EXAMPLES:
        raw_shared = EXAMPLE['shared']
        shareds = [module.Shared.from_dict(raw_shared) for module in modules]
        dumped = [m.dump() for m in shareds]
        assert_all_close(dumped, err_msg='shared_dump')


def test_group():
    for name in MODULES:
        yield _test_group, name


def _test_group(name):
    modules = MODULES[name]
    EXAMPLES = [e for m in modules for e in m.EXAMPLES]
    for EXAMPLE in EXAMPLES:
        values = EXAMPLE['values'][:]
        raw_shared = EXAMPLE['shared']
        temp_shared = modules[0].Shared.from_dict(raw_shared)
        for value in values:
            temp_shared.add_value(value)
        temp_shared.realize()
        raw_shared = temp_shared.dump()

        shareds = [module.Shared.from_dict(raw_shared) for module in modules]
        groups = [
            module.Group.from_values(shared)
            for module, shared in zip(modules, shareds)]
        modules_shareds_groups = zip(modules, shareds, groups)

        for value in values:
            for module, shared, group in modules_shareds_groups:
                group.add_value(shared, value)
            dumped = [g.dump() for g in groups]
            assert_all_close(dumped, err_msg='group_dump')

        for module, shared, group in modules_shareds_groups:
            value = group.sample_value(shared)
            shared.add_value(value)
            values.append(value)

        for value in values:
            scores = [
                group.score_value(shared, value)
                for module, shared, group in modules_shareds_groups
            ]
            assert_all_close(scores, err_msg='score_value')

        scores = [
            group.score_data(shared)
            for module, shared, group in modules_shareds_groups]
        assert_all_close(scores, err_msg='score_data')

        for module, shared, group in modules_shareds_groups:
            dumped = group.dump()
            group.init(shared)
            group.load(dumped)

        scores = [
            group.score_data(shared)
            for module, shared, group in modules_shareds_groups]
        assert_all_close(scores, err_msg='score_data')


def test_plus_group():
    for name in MODULES:
        yield _test_plus_group, name


def _test_plus_group(name):
    modules = MODULES[name]
    modules = [m for m in modules if hasattr(m.Shared, 'plus_group')]
    EXAMPLES = [e for m in modules for e in m.EXAMPLES]
    for EXAMPLE in EXAMPLES:
        raw_shared = EXAMPLE['shared']
        values = EXAMPLE['values']
        shareds = [module.Shared.from_dict(raw_shared) for module in modules]
        groups = [
            module.Group.from_values(shared, values)
            for module, shared in zip(modules, shareds)]
        dumped = [m.plus_group(g).dump() for m, g in zip(shareds, groups)]
        assert_all_close(dumped, err_msg='shared._plus_group(group)')



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from collections import defaultdict
import parsable
from distributions.tests.util import list_models, import_model


@parsable.command
def flavors_by_model():
    '''
    List flavors implemented of each model.
    '''
    models = defaultdict(lambda: [])
    for spec in list_models():
        models[spec['name']].append(spec['flavor'])
    for model in sorted(models):
        print 'model {}: {}'.format(model, ' '.join(sorted(models[model])))


@parsable.command
def models_by_flavor():
    '''
    List models implemented of each flavor.
    '''
    flavors = defaultdict(lambda: [])
    for spec in list_models():
        flavors[spec['flavor']].append(spec['name'])
    for flavor in sorted(flavors):
        print 'flavor {}: {}'.format(flavor, ' '.join(sorted(flavors[flavor])))


@parsable.command
def model_apis():
    '''
    List api of each model.
    '''
    for spec in list_models():
        Model = import_model(spec).Model
        attrs = sorted(attr for attr in dir(Model) if not attr.startswith('_'))
        types = []
        methods = []
        constants = []
        for attr in attrs:
            var = getattr(Model, attr)
            if isinstance(var, type):
                types.append(attr)
            elif hasattr(var, '__call__'):
                methods.append(attr)
            else:
                constants.append(attr)
        print 'distributions.{}.models.{}.{}:'.format(
            spec['flavor'],
            spec['name'],
            Model.__name__)
        print '  types:\n    {}'.format('\n    '.join(types))
        print '  methods:\n    {}'.format('\n    '.join(methods))
        print '  constants:\n    {}'.format('\n    '.join(constants))


if __name__ == '__main__':
    parsable.dispatch()



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from nose.tools import assert_equal
from distributions import fileutil
from distributions import io


EXAMPLES = []
EXAMPLES.append([])
EXAMPLES.append(['asdf'])
EXAMPLES.append([
    {'a': 0, 'b': 'asdf', 'c': [0, 1, 2]},
    {'a': 1, 'b': 'asdf'},
    {'a': 2, 'b': 'asdf'},
    {'a': 3, 'b': 'asdf', 'd': {'inner': 'value'}},
    {'a': 4, 'b': 'asdf'},
    {'a': 5, 'b': 'asdf'},
    {'a': 6, 'b': 'asdf'},
    {'a': 7, 'b': 'asdf'},
    {'a': 8, 'b': 'asdf'},
    {'a': 9, 'b': 'asdf', 'c': [0]},
])


def costream_dump(stream, filename):
    costream = io.stream.json_costream_dump(filename)
    costream.next()
    for item in stream:
        costream.send(item)
    costream.close()


valid_pairs = [
    (io.stream.json_dump, io.stream.json_load),
    (io.stream.json_stream_dump, io.stream.json_load),
    (io.stream.json_stream_dump, io.stream.json_stream_load),
    (costream_dump, io.stream.json_load),
    (costream_dump, io.stream.json_stream_load),
]
named_pairs = {
    (dump.__name__, load.__name__): (dump, load)
    for dump, load in valid_pairs
}


def test_pair():
    for dump, load in named_pairs:
        for filetype in ['', '.gz', '.bz2']:
            yield _test_pair, dump, load, filetype


def _test_pair(dump, load, filetype):
    dump, load = named_pairs[dump, load]
    for example in EXAMPLES:
        print example
        with fileutil.tempdir():
            expected = example
            filename = 'test.json' + filetype
            dump(expected, filename)
            actual = list(load(filename))
            assert_equal(actual, expected)


def test_protobuf_stream():
    for filetype in ['', '.gz', '.bz2']:
        yield _test_protobuf_stream, filetype


def _test_protobuf_stream(filetype):
    filename = 'test.stream' + filetype
    expected = ['asdf', '', 'asdfasdfasdf', 'a', 's', '', '', '', 'd', 'f']
    with fileutil.tempdir():
        print 'dumping'
        io.stream.protobuf_stream_dump(expected, filename)
        print 'loading'
        actual = list(io.stream.protobuf_stream_load(filename))
    assert_equal(actual, expected)



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from libcpp.vector cimport vector

from distributions.rng_cc cimport rng_t
from distributions.lp.vector cimport VectorFloat
from distributions.sparse_counter cimport SparseCounter


ctypedef float Value


cdef extern from "distributions/models/nich.hpp" namespace "distributions::NormalInverseChiSq":
    cppclass Shared:
        float mu
        float kappa
        float sigmasq
        float nu


    cppclass Group:
        int count
        float mean
        float count_times_variance
        void init (Shared &, rng_t &) nogil except +
        void add_value (Shared &, Value &, rng_t &) nogil except +
        void add_repeated_value (Shared &, Value &, int &, rng_t &) nogil except +
        void remove_value (Shared &, Value &, rng_t &) nogil except +
        void merge (Shared &, Group &, rng_t &) nogil except +
        float score_value (Shared &, Value &, rng_t &) nogil except +
        float score_data (Shared &, rng_t &) nogil except +
        Value sample_value (Shared &, rng_t &) nogil except +


    cppclass Sampler:
        void init (Shared &, Group &, rng_t &) nogil except +
        Value eval (Shared &, rng_t &) nogil except +


    cppclass Mixture:
        vector[Group] groups "groups()"
        void init (Shared &, rng_t &) nogil except +
        void add_group (Shared &, rng_t &) nogil except +
        void remove_group (Shared &, size_t) nogil except +
        void add_value \
            (Shared &, size_t, Value &, rng_t &) nogil except +
        void remove_value \
            (Shared &, size_t, Value &, rng_t &) nogil except +
        float score_value_group \
            (Shared &, size_t, Value &, rng_t &) nogil except +
        void score_value \
            (Shared &, Value &, VectorFloat &, rng_t &) nogil except +
        float score_data (Shared &, rng_t &) nogil except +



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

cimport _bnb_h as _h


from libc.stdint cimport uint32_t
from libcpp.vector cimport vector
cimport numpy
numpy.import_array()
from distributions.rng_cc cimport rng_t
from distributions.global_rng cimport get_rng
from distributions.lp.vector cimport (
    VectorFloat,
    vector_float_from_ndarray,
    vector_float_to_ndarray,
)


cdef class Shared:
    cdef _h.Shared * ptr


cdef class Group:
    cdef _h.Group * ptr


cdef class Sampler:
    cdef _h.Sampler * ptr


cdef class Mixture:
    cdef _h.Mixture * ptr
    cdef VectorFloat scores



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy
from distributions.dbg.special import log, gammaln
from distributions.dbg.random import sample_discrete, sample_dirichlet
from distributions.mixins import SharedMixin, GroupIoMixin, SharedIoMixin


NAME = 'DirichletDiscrete'
EXAMPLES = [
    {
        'shared': {'alphas': [0.5] * 4},
        'values': [0, 1, 0, 2, 0, 1, 0],
    },
    {
        'shared': {'alphas': [1.0, 4.0]},
        'values': [0, 1, 1, 1, 1, 0, 1],
    },
    {
        'shared': {'alphas': [2.0 / n for n in xrange(1, 21)]},
        'values': range(20),
    },
]
Value = int


class Shared(SharedMixin, SharedIoMixin):
    def __init__(self):
        self.alphas = None

    @property
    def dim(self):
        return len(self.alphas)

    def load(self, raw):
        self.alphas = numpy.array(raw['alphas'], dtype=numpy.float)

    def dump(self):
        return {'alphas': self.alphas.tolist()}

    def load_protobuf(self, message):
        self.alphas = numpy.array(message.alphas, dtype=numpy.float)

    def dump_protobuf(self, message):
        message.Clear()
        for alpha in self.alphas:
            message.alphas.append(alpha)


class Group(GroupIoMixin):
    def __init__(self):
        self.counts = None

    def init(self, shared):
        self.counts = numpy.zeros(shared.dim, dtype=numpy.int)

    def add_value(self, shared, value):
        self.counts[value] += 1

    def add_repeated_value(self, shared, value, count):
        self.counts[value] += count

    def remove_value(self, shared, value):
        self.counts[value] -= 1

    def merge(self, shared, source):
        self.counts += source.counts

    def score_value(self, shared, value):
        """
        \cite{wallach2009rethinking} Eqn 4.
        McCallum, et. al, 'Rething LDA: Why Priors Matter'
        """
        numer = self.counts[value] + shared.alphas[value]
        denom = self.counts.sum() + shared.alphas.sum()
        return log(numer / denom)

    def score_data(self, shared):
        """
        \cite{jordan2001more} Eqn 22.
        Michael Jordan's CS281B/Stat241B
        Advanced Topics in Learning and Decision Making course,
        'More on Marginal Likelihood'
        """

        dim = shared.dim
        a = shared.alphas
        m = self.counts

        score = sum(gammaln(a[k] + m[k]) - gammaln(a[k]) for k in xrange(dim))
        score += gammaln(a.sum())
        score -= gammaln(a.sum() + m.sum())
        return score

    def sample_value(self, shared):
        sampler = Sampler()
        sampler.init(shared, self)
        return sampler.eval(shared)

    def load(self, raw):
        self.counts = numpy.array(raw['counts'], dtype=numpy.int)

    def dump(self):
        return {'counts': self.counts.tolist()}

    def load_protobuf(self, message):
        self.counts = numpy.array(message.counts, dtype=numpy.int)

    def dump_protobuf(self, message):
        message.Clear()
        for count in self.counts:
            message.counts.append(count)


class Sampler(object):
    def init(self, shared, group=None):
        if group is None:
            self.ps = sample_dirichlet(shared.alphas)
        else:
            self.ps = sample_dirichlet(group.counts + shared.alphas)

    def eval(self, shared):
        return sample_discrete(self.ps)


def sample_group(shared, size):
    group = Group()
    group.init(shared)
    sampler = Sampler()
    sampler.init(shared, group)
    return [sampler.eval(shared) for _ in xrange(size)]
